﻿using System.Xml;
using System.Collections.Generic;
using System;
using System.Collections;
using System.IO;

namespace Persistence {
  public class XmlParser : IEnumerable<IParseToken> {

    private readonly XmlReader reader;
    private delegate IEnumerator<IParseToken> Handler();
    private readonly Dictionary<string, Handler> handlers;

    public XmlParser(TextReader input) {
      XmlReaderSettings settings = new XmlReaderSettings {
                                       ConformanceLevel = ConformanceLevel.Document,
                                       IgnoreWhitespace = true,
                                       IgnoreComments = true
                                     };
      reader = XmlReader.Create(input, settings);
      handlers = new Dictionary<string, Handler> {
                     {"PRIMITIVE", ParsePrimitive},
                     {"COMPOSITE", ParseComposite},
                     {"REFERENCE", ParseReference},
                     {"NULL", ParseNull}
                   };
    }
    public IEnumerator<IParseToken> GetEnumerator() {
      while (reader.Read()) {
        if (!reader.IsStartElement()) {
          break;
        }
        IEnumerator<IParseToken> iterator;
        try {
          iterator = handlers[reader.Name].Invoke();
        }
        catch (KeyNotFoundException) {
          throw new InvalidOperationException(String.Format(
            "No handler for XML tag \"{0}\" installed",
            reader.Name));
        }
        while (iterator.MoveNext()) {
          yield return iterator.Current;
        }
      }
    }
    private IEnumerator<IParseToken> ParsePrimitive() {
      int? id = null;
      string idString = reader.GetAttribute("id");
      if (idString != null)
        id = int.Parse(idString);
      yield return new Primitive(
        reader.GetAttribute("name"),
        Type.GetType(reader.GetAttribute("type")),
        reader.ReadString(),
        id);
    }
    private IEnumerator<IParseToken> ParseComposite() {
      string name = reader.GetAttribute("name");
      Type type = Type.GetType(reader.GetAttribute("type"));
      int id = int.Parse(reader.GetAttribute("id"));
      yield return new CompositeStart(name, type, id);
      IEnumerator<IParseToken> iterator = GetEnumerator();
      while (iterator.MoveNext())
        yield return iterator.Current;
      yield return new CompositeEnd(name, type, id);
    }
    private IEnumerator<IParseToken> ParseReference() {
      yield return new Reference(
        reader.GetAttribute("name"),
        int.Parse(reader.GetAttribute("ref")));
    }
    private IEnumerator<IParseToken> ParseNull() {
      yield return new Null(reader.GetAttribute("name"));
    }
    IEnumerator IEnumerable.GetEnumerator() {
      return GetEnumerator();
    }
  }  
}
